unit main;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, OpenGLContext, Forms, Controls, Graphics,
  Dialogs, ExtCtrls, StdCtrls, gl, glu, Types;


type
  TPaletteItem = record
    Value: GLFloat;
    Color: TColor;
  end;

  TPalette = array of TPaletteItem;

  TCalculateEvent = function(x, y: Double): Double;

  { TForm1 }

  TForm1 = class(TForm)
    cbWireFrame: TCheckBox;
    cbAxes: TCheckBox;
    cbBoundingBox: TCheckBox;
    cbDataPoints: TCheckBox;
    cbFunction: TComboBox;
    Label1: TLabel;
    lblFunction: TLabel;
    OpenGLControl: TOpenGLControl;
    Panel1: TPanel;
    procedure cbAxesChange(Sender: TObject);
    procedure cbBoundingBoxChange(Sender: TObject);
    procedure cbDataPointsChange(Sender: TObject);
    procedure cbFunctionChange(Sender: TObject);
    procedure cbWireFrameChange(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure OpenGLControlMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure OpenGLControlMouseMove(Sender: TObject; Shift: TShiftState;
      X, Y: Integer);
    procedure OpenGLControlMouseWheel(Sender: TObject; Shift: TShiftState;
      WheelDelta: Integer; MousePos: TPoint; var Handled: Boolean);
    procedure OpenGLControlPaint(Sender: TObject);
    procedure OpenGLControlResize(Sender: TObject);
  private
    { private declarations }
    FRotAngleX, FRotAngleY: Double;
    FShiftX, FShiftY: Double;
    FDistance: Double;
    FMouseDownPt: TPoint;
    FPrevMousePt: TPoint;
    FGLInitDone: Boolean;
    FSceneInitDone: boolean;
    FPalette: TPalette;
    FXMin, FXMax, FYMin, FYMax, FZMin, FZMax: GLFloat;
    FNx, FNy: Integer;
    FOnCalculate: TCalculateEvent;

    procedure ApplyProjection;
    procedure DrawAxes;
    procedure DrawBoundingBox;
    procedure DrawDataPoints;
    procedure DrawFunc;

    procedure InitFunc;
    procedure InitGL;
    procedure InitPalette;

    procedure ScreenToWorld(xs, ys, zs: GLFloat; out xw, yw, zw: GLFloat);
    procedure WorldToScreen(xw, yw, zw: GLFloat; out xs, ys, zs: GLFloat);
  end;

var
  Form1: TForm1;

implementation

{$R *.lfm}

uses
  Math;

{ Some demo functions }

function PolynomialFunc(x, y: Double): Double;
begin
  Result := 0.2*(2*x*x*x - x*x - 2*(x+1)*y - 0.5*y*y);
end;


function SincFunc(x, y: Double): Double;
var
  r: Double;
begin
  if (x = 0) and (y = 0) then
    Result := 1
  else begin
    r := sqrt(sqr(x) + sqr(y));
    Result := sin(r) / r;
  end;
end;


function EllipticParaboloid(x, y: Double): Double;
begin
  Result := sqr(x) + sqr(y);
end;


function HyperbolicParaboloid(x, y: Double): Double;
begin
  Result := sqr(x) - sqr(y);
end;


{ Colors }

procedure ClearPalette(var P: TPalette);
begin
  SetLength(P, 0);
end;

// Values must be added in ascending order, this is not checked.
procedure AddToPalette(var P: TPalette; AValue: GLFloat; AColor: TColor);
begin
  SetLength(P, Length(P) + 1);
  P[High(P)].Value := AValue;
  P[High(P)].Color := AColor;
end;

{ borrowed from TAChart }
function InterpolateRGB(AColor1, AColor2: Integer; ACoeff: Double): Integer;
type
  TBytes = packed array [1..4] of Byte;
var
  c1: TBytes absolute AColor1;
  c2: TBytes absolute AColor2;
  r: TBytes absolute Result;
  i: Integer;
begin
  if ACoeff < 0 then ACoeff := 0 else if ACoeff > 1.0 then ACoeff := 1.0;
  for i := 1 to 4 do
    r[i] := Round(c1[i]  + (c2[i] - c1[i]) * ACoeff);
end;

function InterpolateColor(const P: TPalette; AValue: GLFloat): TColor;
var
  i: Integer;
begin
  if AValue < P[0].Value then
    Result := P[0].Color
  else
  if AValue > P[High(P)].Value then
    Result := P[High(P)].Color
  else
    for i:= 0 to High(P) - 1 do
      if (AValue >= P[i].Value) and (AValue <= P[i+1].Value) then begin
        Result := InterpolateRGB(P[i].Color, P[i+1].Color, (AValue - P[i].Value) / (P[i+1].Value - P[i].Value));
        exit;
      end;
end;


{ TForm1 }

procedure TForm1.ApplyProjection;
const
  VIEWANGLE = 40;
  NEAR_CLIPDIST = 0.1;
  FAR_CLIPDIST = 100;
var
  w, h: Integer;
  d: Double;
begin
  w := OpenGLControl.Width;
  h := OpenGLControl.Height;
  if h = 0 then h := 1;

  // Set viewport to window dimensions
  glViewport(0, 0, w, h);

  // Reset coordinate system
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity;

  // Create perspective projection
  gluPerspective(
    VIEWANGLE,      // field of view (degrees) in y direction
    w/h,            // aspect ratio
    NEAR_CLIPDIST,  // distance to near clipping plane
    FAR_CLIPDIST    // distance to far clipping plane
  );

  // Aktuelle Matrix auf ModelView zurücksetzen
  glMatrixMode(GL_MODELVIEW);
end;


procedure TForm1.cbAxesChange(Sender: TObject);
begin
  OpenGLControl.Invalidate;
end;

procedure TForm1.cbBoundingBoxChange(Sender: TObject);
begin
  OpenGLControl.Invalidate;
end;


procedure TForm1.cbDataPointsChange(Sender: TObject);
begin
  OpenGLControl.Invalidate;
end;

procedure TForm1.cbFunctionChange(Sender: TObject);
begin
  case cbFunction.ItemIndex of
    0: begin
         FOnCalculate := @PolynomialFunc;
         FXMin := -2;
         FXMax := +2;
         FYMin := -2;
         FYMax := +2;
       end;
    1: begin
         FOnCalculate := @SincFunc;
         FXMin := -20;
         FXMax := +20;
         FYMin := -20;
         FYMax := +20;
       end;
    2: begin
         FOnCalculate := @EllipticParaboloid;
         FXMin := -3;
         FXMax := +3;
         FYMin := -3;
         FYMax := +3;
       end;
    3: begin
         FOnCalculate := @HyperbolicParaboloid;
         FXMin := -5;
         FXMax := +5;
         FYMin := -5;
         FYMax := +5;
       end;
  end;
  FSceneInitDone := false;
  OpenGLControl.Invalidate;
end;


procedure TForm1.cbWireFrameChange(Sender: TObject);
begin
  FSceneInitDone := false;  // Need a new spacing
  OpenGLControl.Invalidate;
end;


procedure TForm1.DrawAxes;
const
  EPS = 1e-3;
  AXIS_LENGTH = 100;
var
  x0, y0, z0: GLFloat;
begin
  // Calculate position of origin in scaled coordinate system
  WorldToScreen(0, 0, 0, x0, y0, z0);

  glLineWidth(3);
  glEnable(GL_BLEND);  // for line smoothing

  glColor3f(1, 0, 0);   // x axis --> red
  glBegin(GL_LINES);
    glVertex3f(x0, z0 + EPS, y0);
    glVertex3f(x0 + AXIS_LENGTH, z0 + EPS, y0);
  glEnd;

  glColor3f(0, 0, 1);  // OpenGL y axis (= Model z axis) --> blue
  glBegin(GL_LINES);
    glVertex3f(x0, z0 + EPS, y0);
    glVertex3f(x0, z0 + EPS, y0 + AXIS_LENGTH);
  glEnd;

  glColor3f(0, 1, 0);  // OpenGL z axis (= Model y axis) green
  glBegin(GL_LINES);
    glVertex3f(x0, z0 + EPS, y0);
    glVertex3f(x0, z0 + AXIS_LENGTH, y0);
  glEnd;

  glDisable(GL_BLEND);
end;


procedure TForm1.DrawBoundingBox;
var
  x1, x2, y1, y2, z1, z2: GLFloat;
begin
  WorldToScreen(FXMin, FYMin, FZMin, x1, y1, z1);
  WorldToScreen(FXMax, FYMax, FZMax, x2, y2, z2);

  glLineWidth(1);
  glEnable(GL_BLEND);  // for line smoothing

  glColor4f(0, 0, 0, 0.2);
  glBegin(GL_LINE_STRIP);
    glVertex3f(x1, z1, y1);
    glVertex3f(x2, z1, y1);
    glVertex3f(x2, z1, y2);
    glVertex3f(x1, z1, y2);
    glVertex3f(x1, z1, y1);
  glEnd;
  glBegin(GL_LINE_STRIP);
    glVertex3f(x1, z2, y1);
    glVertex3f(x2, z2, y1);
    glVertex3f(x2, z2, y2);
    glVertex3f(x1, z2, y2);
    glVertex3f(x1, z2, y1);
  glEnd;
  glBegin(GL_LINES);
    glVertex3f(x1, z1, y1);
    glVertex3f(x1, z2, y1);
  glEnd;
  glBegin(GL_LINES);
    glVertex3f(x1, z1, y2);
    glVertex3f(x1, z2, y2);
  glEnd;
  glBegin(GL_LINES);
    glVertex3f(x2, z1, y1);
    glVertex3f(x2, z2, y1);
  glEnd;
  glBegin(GL_LINES);
    glVertex3f(x2, z1, y2);
    glVertex3f(x2, z2, y2);
  glEnd;

  glDisable(GL_BLEND);
end;


procedure TForm1.DrawDataPoints;
const
  SIZE = 0.02;
  N = 30;

  procedure DrawNormalizedCube;
  begin
    glBegin(GL_POLYGON);
      glColor3f(1, 0, 0);
      glNormal3f(0.0, 0.0, 1.0);
      glVertex3f(1.0, 1.0, 1.0);
      glVertex3f(-1.0, 1.0, 1.0);
      glVertex3f(-1.0, -1.0, 1.0);
      glVertex3f(1.0, -1.0, 1.0);
    glEnd;

    glBegin(GL_POLYGON);
      glColor3f(0, 1, 0);
      glNormal3f(0.0, 0.0, -1.0);
      glVertex3f(1.0, 1.0, -1.0);
      glVertex3f(1.0, -1.0, -1.0);
      glVertex3f(-1.0, -1.0, -1.0);
      glVertex3f(-1.0, 1.0, -1.0);
    glEnd;

    glBegin(GL_POLYGON);
      glColor3f(0, 0, 1);
      glNormal3f(-1.0, 0.0, 0.0);
      glVertex3f(-1.0, 1.0, 1.0);
      glVertex3f(-1.0, 1.0, -1.0);
      glVertex3f(-1.0, -1.0, -1.0);
      glVertex3f(-1.0, -1.0, 1.0);
    glEnd;

    glBegin(GL_POLYGON);
      glColor3f(1, 1, 0);
      glNormal3f(1.0, 0.0, 0.0);
      glVertex3f(1.0, 1.0, 1.0);
      glVertex3f(1.0, -1.0, 1.0);
      glVertex3f(1.0, -1.0, -1.0);
      glVertex3f(1.0, 1.0, -1.0);
    glEnd;

    glBegin(GL_POLYGON);
      glColor3f(1, 0, 1);
      glNormal3f(0.0, 1.0, 0.0);
      glVertex3f(-1.0, 1.0, -1.0);
      glVertex3f(-1.0, 1.0, 1.0);
      glVertex3f(1.0, 1.0, 1.0);
      glVertex3f(1.0, 1.0, -1.0);
    glEnd;

    glBegin(GL_POLYGON);
      glColor3f(0, 1, 1);
      glNormal3f(0.0, -1.0, 0.0);
      glVertex3f(-1.0, -1.0, -1.0);
      glVertex3f(1.0, -1.0, -1.0);
      glVertex3f(1.0, -1.0, 1.0);
      glVertex3f(-1.0, -1.0, 1.0);
    glEnd;
  end;

  procedure DrawDataPoint(x, y, z: GLFloat);
  var
    xs, ys, zs: GLFloat;
  begin
    WorldToScreen(x, y, z, xs, ys, zs);
    glPushMatrix;
      glTranslatef(xs, zs, ys);
      glScalef(SIZE, SIZE, SIZE);
      DrawNormalizedCube;
    glPopMatrix;
  end;

var
  i: Integer;
  x, y, z: GLFloat;
begin
  if not Assigned(FOnCalculate) then
    exit;

  glPolygonMode(GL_FRONT, GL_FILL);
  glPolygonMode(GL_BACK, GL_FILL);
  glLineWidth(1);

  RandSeed := 0;
  for i := 1 to N do begin
    x := Random * (FXMax - FXMin) + FXMin;
    y := Random * (FYMax - FYMin) + FYMin;
    z := FOnCalculate(x, y) + (Random * 2 - 1) * (FZMax - FZMin)*0.05;   // 5% scatter
    DrawDataPoint(x, y, z);
  end;
end;


procedure TForm1.DrawFunc;

  procedure AddVertex(x, y: GLFloat);
  var
    c: TColor;
    z: GLFloat;
    xs, ys, zs: GLFloat;
  begin
    // Calculate function value
    z := FOnCalculate(x, y);
    // Normalize x, y, z in order to avoid viewing parameters to depend on function
    WorldToScreen(x, y, z, xs, ys, zs);
    // Interpolate color value for z
    c := InterpolateColor(FPalette, zs + 0.5);   // need z = 0 .. 1
    glColor3f(Red(c)/255, Green(c)/255, Blue(c)/255);
    glVertex3f(xs, zs, ys);
    // y, z are exchanged because in OpenGL y points up and z into screen.
  end;

var
  x1, x2, y1, y2, dx, dy: GLFloat;
begin
  if not Assigned(FOnCalculate) then
    exit;

  if cbWireFrame.Checked then begin
    glLineWidth(1);
    glPolygonMode(GL_FRONT, GL_LINE);
    glPolygonMode(GL_BACK, GL_LINE);
  end else begin
    glPolygonMode(GL_FRONT, GL_FILL);
    glPolygonMode(GL_BACK, GL_FILL);
  end;

  dx := (FXMax - FXMin) / (FNx - 1);
  dy := (FYMax - FYMin) / (FNy - 1);

  y1 := FYMin;
  y2 := y1 + dy;
  while y2 <= FYMax do begin
    x1 := FXMin;
    x2 := x1 + dx;
    while x2 <= FXMax do begin
      glBegin(GL_QUADS);
        AddVertex(x1, y1);
        AddVertex(x2, y1);
        AddVertex(x2, y2);
        AddVertex(x1, y2);
      glEnd;
      x1 := x2;
      x2 := x1 + dx;
    end;
    y1 := y2;
    y2 := y1 + dy;
  end;
end;

procedure TForm1.FormActivate(Sender: TObject);
begin
  cbFunctionChange(nil);
end;


procedure TForm1.FormCreate(Sender: TObject);
begin
  FDistance := 4;
  FRotAngleX := 20;
  FRotAngleY := 30;
end;


procedure TForm1.InitFunc;
var
  i, j: Integer;
  x, y, z: GLFloat;
begin
  if not Assigned(FOnCalculate) then
    exit;

  if cbWireFrame.Checked then begin
    FNx := 41;
    FNy := 41;
  end else begin
    FNx := 201;
    FNy := 201;
  end;

  FZMin := 1E30;
  FZMax := -1E30;
  for j := 0 to FNy - 1 do begin
    y := FYMin + (FYMax - FYMin) * j / (FNy - 1);
    for i := 0 to FNx - 1 do begin
      x := FXMin + (FXMax - FXMin) * i / (FNx - 1);
      z := FOnCalculate(x, y);
      FZMin := Min(FZMin, z);
      FZMax := Max(FZMax, z);
    end;
  end;
end;


procedure TForm1.InitGL;
begin
  if FGlInitDone then
    exit;

  FGLInitDone := true;

  glClearColor(0.5, 0.5 ,0.5, 1.0);   // sets background color
  glClearDepth(1.0);
  glDepthFunc(GL_LEQUAL);             // the type of depth test to do
  glEnable(GL_DEPTH_TEST);            // enables depth testing
  glShadeModel(GL_SMOOTH);            // enables smooth color shading

  glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
  glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
end;


procedure TForm1.InitPalette;
begin
  ClearPalette(FPalette);
  AddToPalette(FPalette, 0.00, clBlack);
  AddToPalette(FPalette, 0.25, clBlue);
  AddToPalette(FPalette, 0.50, clRed);
  AddToPalette(FPalette, 0.75, clYellow);
  AddToPalette(FPalette, 1.00, clWhite);
end;


procedure TForm1.OpenGLControlMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  FMouseDownPt := Point(X, Y);
  FPrevMousePt := FMouseDownPt;
end;


procedure TForm1.OpenGLControlMouseMove(Sender: TObject; Shift: TShiftState;
  X, Y: Integer);
const
  SPEED = 0.5;
  DELTA = 0.003;
var
  f: Double;
begin
  f := 1.0;
  if ssShift in Shift then
    f := f * 5.0;
  if ssLeft in Shift then begin
    if FMouseDownPt.Y > OpenGLControl.Height div 2 then begin
      FRotAngleY := FRotAngleY + (X - FPrevMousePt.X)*SPEED*f;
      FRotAngleX := FRotAngleX + (Y - FPrevMousePt.Y)*SPEED*f;
    end
    else begin
      FRotAngleY := FRotAngleY - (X - FPrevMousePt.X)*SPEED*f;
      FRotAngleX := FRotAngleX - (Y - FPrevMousePt.Y)*SPEED*f
    end;
    FPrevMousePt := Point(X, Y);
    OpenGLControl.Invalidate;
  end else
  if ssRight in Shift then begin
    FShiftX := FShiftX + (x - FPrevMousePt.X)*DELTA*f;
    FShiftY := FShiftY + (FPrevMousePt.Y - y)*DELTA*f;
    FPrevMousePt := Point(X, Y);
    OpenGLControl.Invalidate;
  end;
end;


procedure TForm1.OpenGLControlMouseWheel(Sender: TObject;
  Shift: TShiftState; WheelDelta: Integer; MousePos: TPoint;
  var Handled: Boolean);
const
  SPEED = 0.0005;
var
  f: Double;
begin
  if (ssCtrl in Shift) then
    f := -1.0
  else
    f := 1.0;
  if (ssShift in Shift) then
    f := f * 2;
  FDistance := FDistance * (1.0 + f * WheelDelta * SPEED);
  OpenGLControl.Invalidate;
end;


procedure TForm1.OpenGLControlPaint(Sender: TObject);
begin
  if not OpenGLControl.MakeCurrent then
    exit;

  if not FSceneInitDone then begin
    InitPalette;
    InitFunc;
    InitGL;
    ApplyProjection;
    FSceneInitDone := true;
  end;

  glMatrixMode(GL_MODELVIEW);

  glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);
  glLoadIdentity;             { clear the modelview matrix }

  { viewing transformation }
  glTranslatef(FDistance*FShiftX, FDistance*FShiftY, -FDistance);
  glRotatef(FRotAngleX, 1, 0, 0);
  glRotatef(FRotAngleY, 0, 1, 0);

  // Function
  DrawFunc;

  // Data points
  if cbDataPoints.Checked then DrawDataPoints;

  // Bounding box
  if cbBoundingBox.Checked then DrawBoundingBox;

  // Axes
  if cbAxes.Checked then DrawAxes;

  OpenGLControl.SwapBuffers;
end;


procedure TForm1.OpenGLControlResize(Sender: TObject);
begin
  if FSceneInitDone and OpenGLControl.MakeCurrent then
    ApplyProjection;
end;


procedure TForm1.ScreenToWorld(xs, ys, zs: GLFloat; out xw, yw, zw: GLFloat);
begin
  xw := (xs + 1) * (FXMax - FXMin) + FXMin;
  yw := (ys + 1) * (FYMax - FYMin) + FYMin;
  zw := (zs + 0.5) * (FZMax - FZMin) + FZMin;
end;


procedure TForm1.WorldToScreen(xw, yw, zw: GLFloat; out xs, ys, zs: GLFloat);
begin
  xs := (xw - FXMin) / (FXMax - FXMin) * 2 - 1;    // -1 ... +1
  ys := (yw - FYMin) / (FYMax - FYMin) * 2 - 1;    // -1 ... +1
  zs := (zw - FZMin) / (FZMax - FZMin) - 0.5;      // -0.5 ... 0.5
end;


end.

